@@ -17,27 +17,28 @@ and quarantining. Both settings are independent.
 
 Interpretation of @lookup_sql_dsn and @storage_sql_dsn lists is as follows:
 - empty list disables the function and is a default;
-- if both lists are empty no SQL support code will be compiled-in on startup,
+- if both lists are empty no SQL support code will be compiled-in,
   reducing the amount of virtual memory needed for each child process;
 - a list can contain one or more triples: [dsn,user,passw]; more than one
-  triple may be specified to list multiple (backup) SQL servers - the first
-  that responds will be used;
+  triple may be specified to specify multiple (backup) SQL servers - the first
+  that responds will be used as long as it works, then search is retried;
 - if both lists contain refs to the _same_ triples (not just equal triples),
   only one connection to a SQL server will be used; otherwise two independent
   connections to databases will be used, possibly to different SQL servers,
   which may even be of different type (e.g. SQLlite for lookups (read-only),
-  and MySQL for transactional reporting, offering fine lock granularity).
+  and PostgreSQL or MySQL for transactional reporting, offering fine lock
+  granularity).
 
 Example setting:
   @lookup_sql_dsn =
   ( ['DBI:mysql:database=mail;host=127.0.0.1;port=3306', 'user1', 'passwd1'],
     ['DBI:mysql:database=mail;host=host2', 'username2', 'password2'],
-    ['DBI:Pg:host=host1;dbname=mail'],
+    ['DBI:Pg:database=mail;host=host1', 'amavis', '']
     ["DBI:SQLite:dbname=$MYHOME/sql/mail_prefs.sqlite", '', ''] );
 
   @storage_sql_dsn = @lookup_sql_dsn;  # none, same, or separate database
 
-See man page for the Perl module DBI, and corresponding DBD module
+See man page for the Perl module DBI, and corresponding DBD modules
 man pages (DBD::mysql, DBD::Pg, DBD::SQLite, ...) for syntax of the
 first argument.
 
@@ -72,20 +73,24 @@ to create a database. The '--' introduces comments according to SQL specs.
 --     The attribute SERIAL was introduced with MySQL 4.1.0, but it
 --     implicitly creates an additional UNIQUE index, which is redundant.
 --   - instead of declaring a time_iso field in table msgs as a string:
---         time_iso char(16) NOT NULL,
---       it is may be more useful to declare is as:
---         time_iso TIMESTAMP,
---       and set $timestamp_fmt_mysql=1 in amavisd.conf to avoid MySQL
---       inability to accept ISO 8601 timestamps with zone Z and delimiter T
+--       time_iso char(16) NOT NULL,
+--     one may want to declare is as:
+--       time_iso TIMESTAMP NOT NULL DEFAULT 0,
+--     in which case $timestamp_fmt_mysql *MUST* be set to 1 in amavisd.conf
+--     to avoid MySQL inability to accept ISO 8601 timestamps with zone Z
+--     and ISO date/time delimiter T; failing to set $timestamp_fmt_mysql
+--     makes MySQL store zero time on INSERT and write current local time
+--     on UPDATE if auto-update is allowed, which is different from the
+--     intended mail timestamp (localtime vs. UTC, off by seconds)
 
 -- PostgreSQL notes (initially provided by Phil Regnauld):
 --   - use SERIAL instead of INT UNSIGNED NOT NULL AUTO_INCREMENT
 --   - remove the 'unsigned' throughout,
 --   - remove the 'ENGINE=InnoDB' throughout,
---   - instead of declaring a time_iso field in table msgs as a string:
+--   - instead of declaring time_iso field in table msgs as a string:
 --         time_iso char(16) NOT NULL,
---       it is more useful to declare is as:
---         time_iso TIMESTAMP WITHOUT TIME ZONE NOT NULL,
+--       it is more useful to declare it as:
+--         time_iso TIMESTAMP WITH TIME ZONE NOT NULL,
 --       if changing existing table from char to timestamp is desired,
 --       the following clause can be used:
 --         ALTER TABLE msgs ALTER time_iso
@@ -99,10 +104,10 @@ to create a database. The '--' introduces comments according to SQL specs.
 
 -- SQLite notes:
 --   - use INTEGER PRIMARY KEY AUTOINCREMENT
---     instead of INT UNSIGNED NOT NULL AUTO_INCREMENT
+--     instead of INT UNSIGNED NOT NULL AUTO_INCREMENT;
+--   - replace SERIAL by INTEGER;
 --   - SQLite is well suited for lookups database, but is not appropriate
 --     for @storage_sql_dsn due to coarse lock granularity;
---   - replace SERIAL by INTEGER,
 
 -- local users
 CREATE TABLE users (
@@ -214,10 +219,10 @@ CREATE TABLE maddr (
 -- NOTE: records with NULL msgs.content should be ignored by utilities,
 --   as such records correspond to messages just being processes, or were lost
 -- NOTE: with PostgreSQL, instead of a character field time_iso, please use:
---   time_iso timestamp without time zone NOT NULL,
+--   time_iso TIMESTAMP WITH TIME ZONE NOT NULL,
 -- NOTE: with MySQL, instead of a character field time_iso, one might prefer:
---   time_iso timestamp NOT NULL,
---   but a change in the program is required: edit: $timestamp_fmt_mysql=1
+--   time_iso TIMESTAMP NOT NULL DEFAULT 0,
+--   but the following MUST then be set in amavisd.conf: $timestamp_fmt_mysql=1
 CREATE TABLE msgs (
   mail_id    varchar(12)   NOT NULL PRIMARY KEY,  -- long-term unique mail id
   secret_id  varchar(12)   DEFAULT '',  -- authorizes release of mail_id
@@ -236,7 +241,7 @@ CREATE TABLE msgs (
                                         --  none/file/zipfile/bsmtp/sql/mailbox
   quar_loc   varchar(255)  DEFAULT '',  -- quarantine location (e.g. file)
   dsn_sent   char(1),                   -- was DSN sent? Y/N/q (q=quenched)
-  spam_level float,                     -- base message spam level (no boosts)
+  spam_level float,                     -- SA spam level (no boosts)
   message_id varchar(255)  DEFAULT '',  -- mail Message-ID header field
   from_addr  varchar(255)  DEFAULT '',  -- mail From header field,    UTF8
   subject    varchar(255)  DEFAULT '',  -- mail Subject header field, UTF8
@@ -251,7 +256,6 @@ CREATE INDEX msgs_idx_time_num ON msgs (time_num);
 -- per-recipient information related to each processed message;
 -- NOTE: records in msgrcpt without corresponding msgs.mail_id record are
 --  orphaned and should be ignored and eventually deleted by external utilities
--- NOTE: time_num field was introduced to this table with version 2.4.0
 CREATE TABLE msgrcpt (
   mail_id    varchar(12)   NOT NULL,     -- (must allow duplicates)
   rid        integer unsigned NOT NULL,  -- recipient: maddr.id (dupl. allowed)
@@ -261,7 +265,7 @@ CREATE TABLE msgrcpt (
   bl         char(1)       DEFAULT ' ',  -- sender blacklisted by this recip
   wl         char(1)       DEFAULT ' ',  -- sender whitelisted by this recip
   bspam_level float,                     -- spam level + per-recip boost
-  smtp_resp  varchar(255)  DEFAULT '',
+  smtp_resp  varchar(255)  DEFAULT '',   -- SMTP response given to MTA
   FOREIGN KEY (rid)     REFERENCES maddr(id)     ON DELETE RESTRICT,
   FOREIGN KEY (mail_id) REFERENCES msgs(mail_id) ON DELETE CASCADE
 ) ENGINE=InnoDB;
@@ -271,7 +275,6 @@ CREATE INDEX msgrcpt_idx_rid      ON msgrcpt (rid);
 -- mail quarantine in SQL, enabled by $*_quarantine_method='sql:'
 -- NOTE: records in quarantine without corresponding msgs.mail_id record are
 --  orphaned and should be ignored and eventually deleted by external utilities
--- NOTE: time_num field was introduced to this table with version 2.4.0
 CREATE TABLE quarantine (
   mail_id    varchar(12)   NOT NULL,    -- long-term unique mail id
   chunk_ind  integer unsigned NOT NULL, -- chunk number, starting with 1
@@ -281,7 +284,7 @@ CREATE TABLE quarantine (
 ) ENGINE=InnoDB;
 
 -- field msgrcpt.rs is primarily intended for use by quarantine management
--- software; the value initially assigned by amavisd is a space;
+-- software; the value assigned by amavisd is a space;
 -- a short _preliminary_ list of possible values:
 --   'V' => viewed (marked as read)
 --   'R' => released (delivered) to this recipient
@@ -376,8 +379,9 @@ INSERT INTO wblist VALUES (17, 4, 'B');
 
 
 
-Few examples of a query:
+Some examples of a query:
 
+-- mail from last two minutes (MySQL):
 SELECT
   UNIX_TIMESTAMP()-msgs.time_num AS age, SUBSTRING(policy,1,2) as pb,
   content AS c, dsn_sent as dsn, ds, bspam_level AS level, size,
@@ -387,10 +391,23 @@ SELECT
   FROM msgs LEFT JOIN msgrcpt         ON msgs.mail_id=msgrcpt.mail_id
             LEFT JOIN maddr AS sender ON msgs.sid=sender.id
             LEFT JOIN maddr AS recip  ON msgrcpt.rid=recip.id
-  WHERE content IS NOT NULL AND UNIX_TIMESTAMP()-msgs.time_num < 100
-  ORDER BY msgs.time_num DESC LIMIT 15;
+  WHERE content IS NOT NULL AND UNIX_TIMESTAMP()-msgs.time_num < 120
+  ORDER BY msgs.time_num DESC;
 
-# clean messages ordered by count, grouped by domain:
+-- mail from last two minutes (PostgreSQL):
+SELECT
+  now()-time_iso AS age, SUBSTRING(policy,1,2) as pb,
+  content AS c, dsn_sent as dsn, ds, bspam_level AS level, size,
+  SUBSTRING(sender.email,1,18) AS s,
+  SUBSTRING(recip.email,1,18)  AS r,
+  SUBSTRING(msgs.subject,1,10) AS subj
+  FROM msgs LEFT JOIN msgrcpt         ON msgs.mail_id=msgrcpt.mail_id
+            LEFT JOIN maddr AS sender ON msgs.sid=sender.id
+            LEFT JOIN maddr AS recip  ON msgrcpt.rid=recip.id
+  WHERE content IS NOT NULL AND now() - time_iso < INTERVAL '2 minutes'
+  ORDER BY msgs.time_num DESC;
+
+-- clean messages ordered by count, grouped by domain:
 SELECT count(*) as cnt, avg(bspam_level), sender.domain
   FROM msgs
   LEFT JOIN msgrcpt ON msgs.mail_id=msgrcpt.mail_id
@@ -399,21 +416,24 @@ SELECT count(*) as cnt, avg(bspam_level), sender.domain
   WHERE content='C'
   GROUP BY sender.domain ORDER BY cnt DESC LIMIT 50;
 
-# most spamy domains, sorted by spam average, grouped by domain:
+-- top spamy domains with >10 messages, sorted by spam average,
+-- grouped by domain:
 SELECT count(*) as cnt, avg(bspam_level) as spam_avg, sender.domain
   FROM msgs
   LEFT JOIN msgrcpt ON msgs.mail_id=msgrcpt.mail_id
   LEFT JOIN maddr AS sender ON msgs.sid=sender.id
   LEFT JOIN maddr AS recip ON msgrcpt.rid=recip.id
-  GROUP BY sender.domain ORDER BY spam_avg DESC LIMIT 50;
+  WHERE bspam_level IS NOT NULL
+  GROUP BY sender.domain HAVING count(*) > 10
+  ORDER BY spam_avg DESC LIMIT 50;
 
-# sender domains with >100 messages, sorted on sender.domain:
+-- sender domains with >100 messages, sorted on sender.domain:
 SELECT count(*) as cnt, avg(bspam_level) as spam_avg, sender.domain
   FROM msgs
   LEFT JOIN msgrcpt ON msgs.mail_id=msgrcpt.mail_id
   LEFT JOIN maddr AS sender ON msgs.sid=sender.id
   LEFT JOIN maddr AS recip ON msgrcpt.rid=recip.id
-  GROUP BY sender.domain HAVING cnt > 100
+  GROUP BY sender.domain HAVING count(*) > 100
   ORDER BY sender.domain DESC LIMIT 100;
 
 
@@ -444,33 +464,69 @@ SQL schema to the 2.4.0 schema:
     ADD FOREIGN KEY (rid) REFERENCES maddr(id) ON DELETE RESTRICT;
 
 
-An example of a log/report/quarantine database housekeeping;
+BRIEF MySQL EXAMPLE of a log/report/quarantine database housekeeping
+====================================================================
+
+DELETE FROM msgs WHERE time_num < UNIX_TIMESTAMP() - 14*24*60*60;
+DELETE FROM msgs WHERE time_num < UNIX_TIMESTAMP() - 60*60 AND content IS NULL;
+DELETE FROM maddr
+  WHERE NOT EXISTS (SELECT 1 FROM msgs    WHERE sid=id)
+    AND NOT EXISTS (SELECT 1 FROM msgrcpt WHERE rid=id);
+
+
+BRIEF MySQL EQUIVALENT EXAMPLE based on time_iso if its data type is TIMESTAMPS
+===============================================================================
+(don't forget to set: $timestamp_fmt_mysql=1 in amavisd.conf)
+
+DELETE FROM msgs WHERE time_iso < UTC_TIMESTAMP() - INTERVAL 14 day;
+DELETE FROM msgs WHERE time_iso < UTC_TIMESTAMP() - INTERVAL 1 hour
+  AND content IS NULL;
+DELETE FROM maddr
+  WHERE NOT EXISTS (SELECT 1 FROM msgs    WHERE sid=id)
+    AND NOT EXISTS (SELECT 1 FROM msgrcpt WHERE rid=id);
+
+
+BRIEF PostgreSQL EXAMPLE of a log/report/quarantine database housekeeping
+=========================================================================
+
+DELETE FROM msgs WHERE time_iso < now() - INTERVAL '14 days';
+DELETE FROM msgs WHERE time_iso < now() - INTERVAL '1 h' AND content IS NULL;
+DELETE FROM maddr
+  WHERE NOT EXISTS (SELECT 1 FROM msgs    WHERE sid=id)
+    AND NOT EXISTS (SELECT 1 FROM msgrcpt WHERE rid=id);
+
+
+COMMENTED LONGER EXAMPLE of a log/report/quarantine database housekeeping
+=========================================================================
 
 --  discarding indexes makes deletion faster; if we expect a large proportion
 --  of records to be deleted it may be quicker to discard index, do deletions,
---  and re-create index; for daily maintenance this does not pay off
+--  and re-create index (not necessary with PostgreSQL, may benefit MySQL);
+--  for daily maintenance this does not pay off
 --DROP INDEX msgs_idx_sid         ON msgs;
 --DROP INDEX msgrcpt_idx_rid      ON msgrcpt;
 --DROP INDEX msgrcpt_idx_mail_id  ON msgrcpt;
 
 --  delete old msgs records based on timestamps only (for time_iso see next),
 --  and delete leftover msgs records from aborted mail checking operations
-DELETE FROM msgs WHERE time_num < UNIX_TIMESTAMP()-7*24*60*60;
+DELETE FROM msgs WHERE time_num < UNIX_TIMESTAMP()-14*24*60*60;
 DELETE FROM msgs WHERE time_num < UNIX_TIMESTAMP()-60*60 AND content IS NULL;
 
---  instead of purging based on numerical Unix timestamp as above, one may
---  select records based on ISO 8601 timestamps. This is particularly suitable
---  for PostgreSQL:
---DELETE FROM msgs WHERE time_iso < now() - interval '7 days';
---DELETE FROM msgs WHERE time_iso < now() - interval '1 h' AND content IS NULL;
+--  provided the time_iso field was created as type TIMESTAMP DEFAULT 0 (MySQL)
+--  or TIMESTAMP WITH TIME ZONE (PostgreSQL), instead of purging based on
+--  numerical Unix timestamp as above, one may select records based on ISO 8601
+--  UTC timestamps. This is particularly suitable for PostgreSQL:
+--DELETE FROM msgs WHERE time_iso < now() - INTERVAL '14 days';
+--DELETE FROM msgs WHERE time_iso < now() - INTERVAL '1 h' AND content IS NULL;
 and is also possible with MySQL, using slightly different format:
---DELETE FROM msgs WHERE time_iso < now() - interval 7 day;
---DELETE FROM msgs WHERE time_iso < now() -interval 1 hour AND content IS NULL;
-
+--DELETE FROM msgs
+--  WHERE time_iso < UTC_TIMESTAMP() - INTERVAL 14 day;
+--DELETE FROM msgs
+--  WHERE time_iso < UTC_TIMESTAMP() - INTERVAL 1 hour AND content IS NULL;
 
 --  optionally certain content types may be given shorter lifetime
-DELETE FROM msgs WHERE time_num < UNIX_TIMESTAMP()-7*24*60*60
-  AND (content='V' OR (content='S' AND spam_level>20));
+--DELETE FROM msgs WHERE time_num < UNIX_TIMESTAMP()-7*24*60*60
+--  AND (content='V' OR (content='S' AND spam_level>20));
 
 --  (optional) just in case the ON DELETE CASCADE did not do its job, we may
 --  explicitly delete orphaned records (with no corresponding msgs entry);
@@ -490,5 +546,5 @@ DELETE FROM maddr
   WHERE NOT EXISTS (SELECT 1 FROM msgs    WHERE sid=id)
     AND NOT EXISTS (SELECT 1 FROM msgrcpt WHERE rid=id);
 
---  (optional) optimize tables from time to time
+--  (optional) optimize tables once in a while
 --OPTIMIZE TABLE msgs, msgrcpt, quarantine, maddr;
